#include <compare>
#include <iostream>
#include <set>

struct rev_lex {
	int x, y;

	friend bool operator==(const rev_lex&, const rev_lex&) = default;

	friend auto operator<=>(const rev_lex& a,
	                        const rev_lex& b) {
		if (auto cmp = a.y <=> b.y; cmp != 0) {
			return cmp;
		}
		return a.x <=> b.x;
	}

	void show() const {
		std::cout << "{ " << x << ", " << y << " }";
	}
};

// Python-like set comparison (using <, <=, >, >= for set inclusion)
// to demonstrate std::partial ordering.
// Note that in this case, implementing <, > separately can be more efficient.
// (Think about it!)
struct my_set {
	std::set<int> s;

	void show() const {
		if (s.empty()) {
			std::cout << "{}";
			return;
		}
		char sep = '{';
		for (int x : s) {
			std::cout << sep << ' ' << x;
			sep = ',';
		}
		std::cout << " }";
	}

	friend bool operator==(const my_set&, const my_set&) = default;

	friend std::partial_ordering operator<=>(const my_set& left,
	                                         const my_set& right) {
		// There are more efficient implementations than this one,
		// but we would need to use iterators.
		std::size_t rcount = 0;
		bool lmore = false;
		for (int lx : left.s) {
			if (right.s.contains(lx)) {
				++rcount;
			} else {
				lmore = true;
			}
		}
		bool rmore = rcount < right.s.size();
		return  lmore &&  rmore ? std::partial_ordering::unordered :
		        lmore && !rmore ? std::partial_ordering::greater   :
		       !lmore &&  rmore ? std::partial_ordering::less      :
		                          std::partial_ordering::equivalent;
	}
};

// A slightly weird fraction type, to demonstrate std::weak_ordering.
struct fraction {
	int num = 0, denom = 1;

	void show() const {
		std::cout << num << '/' << denom;
	}

	friend std::weak_ordering operator<=>(const fraction& a,
	                                      const fraction& b) {
		return a.num * b.denom <=> b.num * a.denom;
	}

	friend bool operator==(const fraction& a, const fraction& b) {
		// We don't have a better implementation of == in this case,
		// so we delegate:
		return (a <=> b) == 0;
	}
};


void show_cmp(const auto& a, const auto& b) {
	std::cout << "-----------------\na = ";
	a.show();
	std::cout << "\nb = ";
	b.show();
	std::cout << "\na < b:  " << (a < b)
	          << "\na <= b: " << (a <= b)
	          << "\na == b: " << (a == b)
	          << "\na != b: " << (a != b)
	          << "\na >= b: " << (a >= b)
	          << "\na > b:  " << (a > b) << '\n';
}

int main() {
	std::cout << "---rev_lex---\n";
	show_cmp(rev_lex{ 1, 2 }, rev_lex{ 3, 1 });
	std::cout << "----my_set---\n";
	show_cmp(my_set{{ 1, 2, 3 }}, my_set{{ 3, 2, 1 }});
	show_cmp(my_set{{ 17, 42 }}, my_set{});
	show_cmp(my_set{{ 5, 10, 12 }}, my_set{{ 17, 5, 0 }});
	std::cout << "---fraction--\n";
	show_cmp(fraction{}, fraction{3, 4});
	show_cmp(fraction{1, 1}, fraction{3, 4});
	show_cmp(fraction{9, 12}, fraction{3, 4});

	std::set<fraction> fs = { {1, 2}, {3, 4}, {6, 8}, {-7, 4} };
	fs.emplace(28, 16);
	fs.emplace(7, 4);
	fs.emplace(10, 20);
	for (auto f : fs) {
		std::cout << ' ';
		f.show();
	}
	std::cout << '\n';
}
